#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <math.h>
#include <omp.h>
#include <cblas.h>
#include <cuda_runtime.h>
#include <cublas_v2.h>

#define	ITER	1

#if TYPE_DOUBLE
#define	TYPE	double
#else
#define	TYPE	float
#endif

#define USE_CUDA_EVENT_TIMER	0
#define CHECK_ANSWER		1
#include "test-common.h"
#include "test.h"

// macro for the legacy cublas
#define	cublasAlloc(a,b,c)	cudaMalloc( (c),(a)*(b) )
#define	cublasFree(a)		cudaFree( a )


int
set_matrix_vectors( const int N, TYPE *A, const int lda, TYPE *X, TYPE *Y,
                    const enum
#if __isDOUBLE__
                    D_FUNCS
#else
                    S_FUNCS
#endif
                    func )
{

  srand(0);

  for(int i=0;i<N;i++) {
    for(int j=0;j<N;j++) {
      A[j+i*lda]=(TYPE)(0);
    }
  }

  for(int i=0;i<N;i++) X[i]=(TYPE)1;
  for(int i=0;i<N;i++) Y[i]=(TYPE)0;

  for(int i=0;i<N;i++) {
    for(int j=0;j<N;j++) {
      TYPE  t  = (TYPE)rand();
      A[j+lda*i] = t / RAND_MAX;
    }
  }

  if ( func == SYMV_U ) {
    for(int i=0;i<N;i++) {
      for(int j=0;j<i;j++) {
        TYPE  t1  = A[j+lda*i];
        TYPE  t2  = A[i+lda*j];
        A[i+lda*j] = (TYPE)0;
        A[j+lda*i] = (t1+t2)/2;
      }
    }
  }
  if ( func == SYMV_L ) {
    for(int i=0;i<N;i++) {
      for(int j=i+1;j<N;j++) {
        TYPE  t1  = A[j+lda*i];
        TYPE  t2  = A[i+lda*j];
        A[i+lda*j] = (TYPE)0;
        A[j+lda*i] = (t1+t2)/2;
      }
    }
  }
}

int
main(int argc, char *argv[])
{
  TYPE	alpha = (TYPE)3.0;
  TYPE	beta  = (TYPE)0.0;//4.0;
  TYPE	*A, *B, *C, *D;
  TYPE	*dA, *dB, *dC, *dD;
  int	N, M, K;
  int	LDA, LDB, LDC;
  int	i;
  int	blk;
  FILE	*fp;
#if CHECK_ANSWER
  int	verify = 0;
#endif
  int	id = get_gpu_id();


#if __isDOUBLE__
  enum D_FUNCS	func = GEMV_N;
#else
  enum S_FUNCS	func = GEMV_N;
#endif

  char	*func_name = "";

  cudaError_t st;

  if ( cudaSuccess != (st = cudaSetDevice(id)) ) {
    exit(1);
  }

  cublasStatus_t stat;
  cublasHandle_t handle;
  stat = cublasCreate( &handle );

  ASPEN_init( id );

  {
    char *argv0 = argv[0];
    if ( strrchr(argv0, '/') != argv0 ) {
      argv0 = strrchr(argv0,'/')+1;
    }
    if ( 0 == strcmp(argv0,TEST_GEMV_N) ) {
      func      = GEMV_N;
      func_name = NAME_GEMV_N;
    }
    if ( 0 == strcmp(argv0,TEST_GEMV_T) ) {
      func      = GEMV_T;
      func_name = NAME_GEMV_T;
    }
    if ( 0 == strcmp(argv0,TEST_SYMV_U) ) {
      func      = SYMV_U;
      func_name = NAME_SYMV_U;
    }
    if ( 0 == strcmp(argv0,TEST_SYMV_L) ) {
      func      = SYMV_L;
      func_name = NAME_SYMV_L;
    }
  }

  if ( argc > 1 ) {
    fp=fopen(argv[1], "r");
  } else {
    fp=fopen("IN", "r");
  }
  if ( fp == NULL ) {
    fprintf(stderr,"Cannot open parameter file.\n");
    exit(1);
  }

  fscanf(fp,"%d", &N);

  if ( argc > 2 ) {
    sscanf(argv[2], "%d", &blk);
    printf("% BLK specified=%d\n", blk);
  } else {
    blk = -1; //64;
  }

  {
    verify = 1;
  }

  print_head( func_name );

#if CHECK_ANSWER
  printf("Run in the %s mode\n",
         verify ? "verify" : "non-verify(fast)" );
  fflush(stdout);
#endif
  printf("*****************************************/-->\n");
  fflush(stdout);


  int N_fault=0;

#if 0
  warm_up_GPU( handle );
#endif

  for(int Itr = 0; fscanf(fp,"%d", &N) > 0; Itr++ ) {

    if ( N <= 1 ) break;

    LDA = ((N-1)/(256/sizeof(TYPE))+1);
    LDA *= (256/sizeof(TYPE));

    int	M = N;
    int	LDM = ((M-1)/(256/sizeof(TYPE))+1);
    LDM *= (256/sizeof(TYPE));

    //	cudaMallocHost((void **)&A, sizeof(TYPE) *(N*LDA));
    cudaMallocHost((void **)&A, sizeof(TYPE) *(M*LDM));
    cudaMallocHost((void **)&B, sizeof(TYPE) *(N*1));
    cudaMallocHost((void **)&C, sizeof(TYPE) *(N*1));
    cudaMallocHost((void **)&D, sizeof(TYPE) *(N*1));

    //	cublasAlloc(N*LDA, sizeof(TYPE), (void**)&dA);
    cublasAlloc(M*LDM, sizeof(TYPE), (void**)&dA);
    cublasAlloc(N*1, sizeof(TYPE), (void**)&dB);
    cublasAlloc(N*1, sizeof(TYPE), (void**)&dC);
    cublasAlloc(N*1, sizeof(TYPE), (void**)&dD);

    if(dA==NULL||dB==NULL||dC==NULL||dD==NULL) {
      fprintf(stderr, "# Fail to allocate Dev_array %dx%d, %d*4\n",
              N, LDA, N);
      fprintf(stderr, "# Fail to allocate memory %x %x %x %x\n",
              dA, dB, dC, dD);
      if(dA!=NULL) cublasFree(dA);
      if(dB!=NULL) cublasFree(dB);
      if(dC!=NULL) cublasFree(dC);
      if(dD!=NULL) cublasFree(dD);
      N_fault++;
      goto FFF;
    } else {
      N_fault=0;
    }

    printf("N= %d ", N);
    fflush(stdout);
    fflush(stdout);
    fflush(stdout);


    set_matrix_vectors( N, A, LDA, B, C, func );

    for(int i=0;i<N;i++) D[i] = C[i];


    if ( argc > 3 ) { ; } else {
      cublasSetMatrix(LDM, M, sizeof(TYPE), A, LDM, dA, LDM);
      cublasSetVector(N, sizeof(TYPE), B, 1, dB, 1);
      cublasSetVector(N, sizeof(TYPE), C, 1, dC, 1);
      cublasSetVector(N, sizeof(TYPE), D, 1, dD, 1);
    }

    fflush(stdout);
    cudaThreadSynchronize();

    double	tt[1000];
    double	t1 = get_wtime();

    for(i=0;i<ITER;i++){

#if USE_CUDA_EVENT_TIMER
      cudaEvent_t	prestart, start, stop;
      float	elapsed_time_ms;

      cudaEventCreate( &prestart );
      cudaEventCreate( &start );
      cudaEventCreate( &stop );
#endif

#if !USE_CUDA_EVENT_TIMER
      cudaThreadSynchronize();
      double	tt1 = get_wtime();
#endif

      int	incb = 1;
      int	incc = 1;
      char	*opt = "U";

#if USE_CUDA_EVENT_TIMER
      cudaEventRecord( prestart, 0 );
      cudaEventSynchronize( prestart );
      cudaEventRecord( start, 0 );
#endif

      switch (func) {
      case GEMV_N:
        ASPEN_GEMV_n (blk, alpha, beta, dA, dB, dC, LDA, N, N);
        break;
      case GEMV_T:
        ASPEN_GEMV_t (blk, alpha, beta, dA, dB, dC, LDA, N, N);
        break;
      case SYMV_U:
        ASPEN_SYMVu (blk, N, dA, LDA, dB, dC, alpha, beta);
        break;
      case SYMV_L:
        ASPEN_SYMVl (blk, N, dA, LDA, dB, dC, alpha, beta);
        break;
      }

#if 1
      cudaError_t cu_err = cudaGetLastError();
      if ( cu_err != cudaSuccess ) {
        printf("CUDA internal ERROR detected\n"); goto HHH;
      }
#endif

#if USE_CUDA_EVENT_TIMER
      cudaEventRecord( stop, 0 );
      cudaEventSynchronize( stop );
      cudaEventElapsedTime( &elapsed_time_ms, start, stop );
      tt[i]=((double)elapsed_time_ms)*1e-3;
#else
      cudaThreadSynchronize();
      double	tt2 = get_wtime();
      tt[i]=tt2-tt1;
#endif

      if ( argc > 3 ) goto FIN;

#if CHECK_ANSWER
      if(verify){
        if(i==0&&N<50000){
          cudaThreadSynchronize();
          do_usleep(100);
          for(int j=0;j<N;j++) { C[j] = (TYPE)0; }
          cublasGetVector(N, sizeof(TYPE), dC, 1, C, 1);
          cudaThreadSynchronize();
          do_usleep(100);
#if 0
          switch (func) {
          case GEMV_N:
            cublasgemv
              ( handle, CUBLAS_OP_N, N, N, &alpha, dA, LDA, dB, 1, &beta, dD, 1);
            break;
          case GEMV_T:
            cublasgemv
              ( handle, CUBLAS_OP_T, N, N, &alpha, dA, LDA, dB, 1, &beta, dD, 1);
            break;
          case SYMV_U:
            cublassymv
              ( handle, CUBLAS_FILL_MODE_UPPER, N, &alpha, dA, LDA, dB, 1, &beta, dD, 1);
            break;
          case SYMV_L:
            cublassymv
              ( handle, CUBLAS_FILL_MODE_LOWER, N, &alpha, dA, LDA, dB, 1, &beta, dD, 1);
            break;
          }
          int	jj = cublasGetError();
          if ( jj != 0 ) {
            printf("BLAS ERROR detected\n"); goto HHH;
          }
          cudaThreadSynchronize();
          do_usleep(100);
          cublasGetVector(N, sizeof(TYPE), dD, 1, D, 1);
#else
          switch (func) {
          case GEMV_N:
            cblas_gemv
              ( CblasColMajor, CblasNoTrans, N, N, alpha, A, LDA, B, 1, beta, D, 1);
            break;
          case GEMV_T:
            cblas_gemv
              ( CblasColMajor, CblasTrans, N, N, alpha, A, LDA, B, 1, beta, D, 1);
            break;
          case SYMV_U:
            cblas_symv
              ( CblasColMajor, CblasUpper, N, alpha, A, LDA, B, 1, beta, D, 1);
            break;
          case SYMV_L:
            cblas_symv
              ( CblasColMajor, CblasLower, N, alpha, A, LDA, B, 1, beta, D, 1);
            break;
          }
#endif
          do_usleep(100);

          int	kk=0;
#pragma omp parallel for reduction(+:kk)
          for(int j=0;j<N;j++){
            // check NAN
            int s = isnan(C[j])||isnan(-C[j]);
            if ( s ) { kk++; }
          }
          if ( kk > 0 ) {
            printf(" NAN detectedd\n");
            goto GGG; }

#if __isDOUBLE__
#define	FABS	fabs
#else
#define	FABS	fabsf
#endif

          TYPE t1 = (TYPE)0;
          TYPE t2 = (TYPE)0;
          TYPE t3 = (TYPE)0;
#pragma omp parallel for reduction(+:t1,t2,t3)
          for(int j=0;j<N;j++){
            t1 += FABS(C[j]-D[j]);
            t2 += FABS(C[j]);
            t3 += FABS(D[j]);
          }

          const TYPE tol = 32*EPS;

          printf( " | ASPEN - CBLAS | = %g\n",         t1     );
          printf( " | ASPEN | = %g, | CBLAS | = %g\n", t2, t3 );
          printf( " TOL = %g ", tol*t3 );

          if ( t1 >= tol * t2 ) {
            printf( " **ERROR** \n" );
          } else {
            printf( " ++PASSED++\n" );
          }

          cudaThreadSynchronize();
          do_usleep(100);
        }
      }
#endif

    }

#if 0
    int status = cublasGetError();
    if (status != CUBLAS_STATUS_SUCCESS) {
      fprintf (stderr, "!!!! kernel execution error.\n");
      goto GGG; //return EXIT_FAILURE;
    }
#endif
  EEE:;

    cudaThreadSynchronize();

    fflush(stdout);
  GGG:
    fflush(stdout);

    cublasFree(dA); //printf(":dA");
    cublasFree(dB); //printf(":dB");
    cublasFree(dC); //printf(":dC");
    cublasFree(dD); //printf(":dC");

  FFF:
    cudaFreeHost(A); //printf(":A");
    cudaFreeHost(B); //printf(":B");
    cudaFreeHost(C); //printf(":C");
    cudaFreeHost(D); //printf(":D");

    printf("\n");
    fflush(stdout);
    fflush(stdout);

    if ( N_fault > 4 ) break;
    if ( feof(fp) ) break;
  }
  cublasDestroy( handle );
  ASPEN_shutdown();
 HHH:
  cudaDeviceReset();
  fclose(fp);
 FIN:
  ;
}

