// additional including standard header files
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#include <ctype.h>
#include <string.h>
#include <stddef.h>
#include <stdint.h>


// additional including CUDA header files
#include <cuda_runtime.h>
#include <cuda.h>
#if CUDA_VERSION<5000
#   include <cutil.h>
#endif
#include <cublas.h>


#ifndef ASPEN_DEVEL_H_INCLUDED

// definitions for generic macros
#define _MACRO_WRAP_(...)                       \
  do { { __VA_ARGS__; } break; } while( 0 )
#define _MACRO_BLOCK_		_MACRO_WRAP_

#define __4PARAMS__(N,x,y,z,w,...)	N##_##x##_##y##_##z##_##w
#define __3PARAMS__(N,x,y,z,...)	N##_##x##_##y##_##z
#define __2PARAMS__(N,x,y,...)		N##_##x##_##y
#define	__1PARAMS__(N,x,...)		N##_##x

#define _NAME_4PARAMS_(N,x,y,z,w,...)	__4PARAMS__(N,x,y,z,w)
#define _NAME_3PARAMS_(N,x,y,z,...)	__3PARAMS__(N,x,y,z)
#define _NAME_2PARAMS_(N,x,y,...)	__2PARAMS__(N,x,y)
#define	_NAME_1PARAMS_(N,x,...)		__1PARAMS__(N,x)

// definitions for the prefix of API
#define	API_PREFIX	ASPEN
#define	API(x,...)	_NAME_1PARAMS_(API_PREFIX,x)



// CUDA architecture CONST
#define	Tesla		(130)
#define	Fermi		(200)
#define	Kepler		(300)
#define	Kepler2		(350)
#define	Maxwell		(500)
#define	Maxwell2	(520)


#define	ASPEN_GPU_Names(x,...)                          \
  (                                                     \
   ( ( x == Tesla )        ? "Tesla"                    \
     : ( ( x == Fermi )    ? "Fermi"                    \
         : ( ( x == Kepler )   ? "Kepler"               \
             : ( ( x == Kepler2 )  ? "Kepler2"          \
                 : ( ( x == Maxwell )  ? "Maxwell"      \
                     : ( ( x == Maxwell2 ) ? "Maxwell2" \
                         : ( "unknown" ) ) ) ) ) ) ) )  \
  )


#if defined(CURRENT_GPU)
#if CURRENT_GPU==Tesla
#	define	GPU_tuned	TESLA
#endif
#if CURRENT_GPU==Fermi
#	define	GPU_tuned	FERMI
#endif
#if CURRENT_GPU==Kepler
#	define	GPU_tuned	KEPLER
#endif
#if CURRENT_GPU==Kepler2
#	define	GPU_tuned	KEPLER2
#endif
#if CURRENT_GPU==Maxwell
#	define	GPU_tuned	MAXWELL
#endif
#if CURRENT_GPU==Maxwell2
#	define	GPU_tuned	MAXWELL2
#endif
#else
#	define	GPU_tuned	UNKNOWN
#endif


//
// The macro __CUDA_ARCH__ is only available on the compilation for devices
// To use __CUDA_ARCH__ sometimes confuses us, so we separately compile
// the source code xxxx_X-template.cu by nvcc option such as
// --generate-code arch=compute_XX,code=compute_XX, and GPU_ARCH can be
// used for the distingishment of the GPU kernel.
//

#if defined(GPU_ARCH)
#if GPU_ARCH == Tesla
// Tesla, GT2xx
#define	GPU_PREFIX		TESLA
#define	MAX_THREAD_BLOCKS       (8)
#define	MAX_REGS_PER_BLOCK	(32768)
#define	SHMEM_UNIT_SIZE		(512)
#define	REGIS_UNIT_SIZE		(1)
#define	WARP_ALLOC_UNIT		(2)
#define	FENCE(...)		(1)
#define	FENCE_BLOCK(...)	(1)
#define	FENCE_SYSTEM(...)	(1)
#endif
#if GPU_ARCH == Fermi
// Fermi
#define	GPU_PREFIX		FERMI
#define	MAX_THREAD_BLOCKS       (8)
#define	MAX_REGS_PER_BLOCK	(32768)
#define	SHMEM_UNIT_SIZE		(128)
#define	REGIS_UNIT_SIZE		(64)
#define	WARP_ALLOC_UNIT		(2)
#define	FENCE(...)		__threadfence( )
#define	FENCE_BLOCK(...)	__threadfence_block( )
#define	FENCE_SYSTEM(...)	__threadfence_system( )
#endif
#if GPU_ARCH == Kepler
// Kepler, GTX680
#define	GPU_PREFIX		KEPLER
#define	MAX_THREAD_BLOCKS       (16)
#define	MAX_REGS_PER_BLOCK	(65536)
#define	SHMEM_UNIT_SIZE		(256)
#define	REGIS_UNIT_SIZE		(256)
#define	WARP_ALLOC_UNIT		(4)
#define	FENCE(...)		__threadfence( )
#define	FENCE_BLOCK(...)	__threadfence_block( )
#define	FENCE_SYSTEM(...)	__threadfence_system( )
#endif
#if GPU_ARCH == Kepler2
// Kepler
#define	GPU_PREFIX		KEPLER2
#define	MAX_THREAD_BLOCKS       (16)
#define	MAX_REGS_PER_BLOCK	(65536)
#define	SHMEM_UNIT_SIZE		(256)
#define	REGIS_UNIT_SIZE		(256)
#define	WARP_ALLOC_UNIT		(4)
#define	FENCE(...)		__threadfence( )
#define	FENCE_BLOCK(...)	__threadfence_block( )
#define	FENCE_SYSTEM(...)	__threadfence_system( )
#endif
#if GPU_ARCH == Maxwell
// Maxwell
#define	GPU_PREFIX		MAXWELL
#define	MAX_THREAD_BLOCKS       (32)
#define	MAX_REGS_PER_BLOCK	(65536)
#define	SHMEM_UNIT_SIZE		(256)
#define	REGIS_UNIT_SIZE		(256)
#define	WARP_ALLOC_UNIT		(4)
#define	FENCE(...)		__threadfence( )
#define	FENCE_BLOCK(...)	__threadfence_block( )
#define	FENCE_SYSTEM(...)	__threadfence_system( )
#endif
#if GPU_ARCH == Maxwell2
// Maxwell2
#define	GPU_PREFIX		MAXWELL2
#define	MAX_THREAD_BLOCKS       (32)
#define	MAX_REGS_PER_BLOCK	(65536)
#define	SHMEM_UNIT_SIZE		(256)
#define	REGIS_UNIT_SIZE		(256)
#define	WARP_ALLOC_UNIT		(4)
#define	FENCE(...)		__threadfence( )
#define	FENCE_BLOCK(...)	__threadfence_block( )
#define	FENCE_SYSTEM(...)	__threadfence_system( )
#endif
#endif  // GPU_ARCH

// kernel name with GPU_name
#define	ASPEN_tuned_kernel_name(x,...)          \
  _NAME_2PARAMS_(API_PREFIX,x,GPU_tuned)



// Error check and safe invocation of the kernel function
#define	ASPEN_KERNEL_INVOKE( kernel_name, ...)				\
  _MACRO_BLOCK_ (							\
                 kernel_name  __VA_ARGS__ ;				\
                 cudaError_t err = cudaGetLastError();			\
                 if ( err != cudaSuccess ) {				\
                   printf ( "ERROR in [%s] : %s \n",                    \
                            #kernel_name, cudaGetErrorString(err) );    \
                   exit ( 1 );                                          \
                   return;						\
                 }							\
                                                                        )
#define __LOOP_DIVIDER__(...)                   \
  asm( "// <- nop ->;" )



#include "aspen_texture.h"
#include "aspen_complex.h"
#include "aspen_const.h"
#if 0
#include "aspen_atomic.h"
#endif
#include "aspen_ld_st.h"
#include "aspen_istypes.h"



#include "aspen_devel_func.h"

/* Utility functions for C APIs */
extern "C"  int     API(get_numbers_MP)      ( void );
extern "C"  void    API(get_device_Name)     ( char * name );
extern "C"  int     API(get_device_Major)    ( void );
extern "C"  int     API(get_device_Minor)    ( void );
extern "C"  int     API(get_device_CC)       ( void );
extern "C"  int     API(get_driver_Version)  ( void );
extern "C"  int     API(get_runtime_Version) ( void );
extern "C"  size_t  API(get_device_Memory)   ( void );
extern "C"  size_t  API(get_device_WorkSize) ( void );
extern "C"  void *  API(get_device_WorkArea) ( void );

#include "aspen.h"

#define ASPEN_DEVEL_H_INCLUDED		1
#endif

