#ifndef ASPEN_TEXTURE_H_INCLUDED

// definition for Texture memory
#if __isDOUBLE__
#define	DEF_TEXTURE( Tex_x, X_, ... )           \
  texture < uint2, 1 > Tex_x;                   \
  __forceinline__ __device__ scalar_t           \
  X_ ( const int i )                            \
  {                                             \
    uint2 v; double V;                          \
    v = tex1Dfetch ( Tex_x, i );                \
    V = __hiloint2double ( v.y, v.x );          \
    return V;                                   \
  }
#endif

#if __isFLOAT__
#define	DEF_TEXTURE( Tex_x, X_, ... )           \
  texture < float, 1 > Tex_x;                   \
  __forceinline__ __device__ scalar_t           \
  X_ ( const int i )                            \
  {                                             \
    float v;                                    \
    v = tex1Dfetch ( Tex_x, i );                \
    return (scalar_t) v;                        \
  }
#endif

#if __isDOUBLE_COMPLEX__
#define	DEF_TEXTURE( Tex_x, X_, ... )           \
  texture < uint4, 1 > Tex_x;                   \
  __forceinline__ __device__ scalar_t           \
  X_ ( const int i )                            \
  {                                             \
    uint4 v; double2 w; cuDoubleComplex V;      \
    v = tex1Dfetch ( Tex_x, i );                \
    w.x = __hiloint2double ( v.y, v.x );        \
    w.y = __hiloint2double ( v.w, v.z );        \
    V = make_cuDoubleComplex( w.x, w.y );       \
    return V;                                   \
  }
#endif

#if __isFLOAT_COMPLEX__
#define	DEF_TEXTURE( Tex_x, X_, ... )           \
  texture < float2, 1 > Tex_x;                  \
  __forceinline__ __device__ scalar_t           \
  X_ ( const int i )                            \
  {                                             \
    float2 v;                                   \
    v = tex1Dfetch ( Tex_x, i );                \
    return (scalar_t) v;                        \
  }
#endif


#if __isDOUBLE__
#define	SETUP_TEXTURE( Tex_x, x, n, offsetX, ... )			\
  _MACRO_BLOCK_ (							\
                 cudaError_t err;                                       \
                 err = cudaBindTexture (                                \
                                        &offsetX, Tex_x,                \
                                        (int2 *)x,                      \
                                        (n)*sizeof( scalar_t ) );       \
                 if ( err != cudaSuccess ) {                            \
                   printf ( "cannot bind to texture :: %d \n", err );   \
                   return;                                              \
                 }                                                      \
                 offsetX /= sizeof( x[0] );                             \
                                                                        )
#endif

#if __isFLOAT__
#define	SETUP_TEXTURE( Tex_x, x, n, offsetX, ... )			\
  _MACRO_BLOCK_ (							\
                 cudaError_t err;                                       \
                 err = cudaBindTexture(                                 \
                                       &offsetX, Tex_x,                 \
                                       (float *)x,                      \
                                       (n)*sizeof( scalar_t ) );        \
                 if ( err != cudaSuccess ) {                            \
                   printf ( "cannot bind to texture :: %d \n", err );   \
                   return;                                              \
                 }                                                      \
                 offsetX /= sizeof( x[0] );                             \
                                                                        )
#endif

#if __isDOUBLE_COMPLEX__
#define	SETUP_TEXTURE( Tex_x, x, n, offsetX, ... )			\
  _MACRO_BLOCK_ (                                                       \
                 cudaError_t err;                                       \
                 err = cudaBindTexture (                                \
                                        &offsetX, Tex_x,                \
                                        (int4 *)x,                      \
                                        (n)*sizeof( scalar_t ) );       \
                 if ( err != cudaSuccess ) {                            \
                   printf ( "cannot bind to texture :: %d \n", err );   \
                   return;                                              \
                 }                                                      \
                 offsetX /= sizeof( x[0] );                             \
                                                                        )
#endif

#if __isFLOAT_COMPLEX__
#define	SETUP_TEXTURE( Tex_x, x, n, offsetX, ... )			\
  _MACRO_BLOCK_ (							\
                 cudaError_t err;                                       \
                 err = cudaBindTexture(                                 \
                                       &offsetX, Tex_x,                 \
                                       (float2 *)x,                     \
                                       (n)*sizeof( scalar_t ) );        \
                 if ( err != cudaSuccess ) {                            \
                   printf ( "cannot bind to texture :: %d \n", err );   \
                   return;                                              \
                 }                                                      \
                 offsetX /= sizeof( x[0] );                             \
                                                                        )
#endif


#define ASPEN_TEXTURE_H_INCLUDED		1
#endif

