#include "aspen_devel.h"


class ASPEN_GPU_info
{
  //
private:
  int   numbers_MPs;
  char * device_Name;
  int   device_Major;
  int   device_Minor;
  int   driver_Version;
  int   runtime_Version;
  size_t  global_memory;
  size_t  lwork;
  void * work;
  //
public:
  virtual
  ~ASPEN_GPU_info ( void )
  {
    cudaFree( work );
    free ( device_Name );
  } // destructor

  ASPEN_GPU_info ( int device_id )
  {
    int dev = device_id;
    struct cudaDeviceProp deviceProp;

    cudaGetDeviceProperties ( &deviceProp, dev );
    numbers_MPs  = deviceProp.multiProcessorCount;
    int len = strlen ( deviceProp.name );
    device_Name = (char *) malloc ( len + 1 );
    strcpy( device_Name, deviceProp.name );
    device_Major = deviceProp.major;
    device_Minor = deviceProp.minor;

    cudaDriverGetVersion( &driver_Version );
    cudaRuntimeGetVersion( &runtime_Version );

    global_memory = deviceProp.totalGlobalMem;
    lwork = global_memory/sizeof(double);
    lwork = (size_t)(sqrt((double)lwork));
    lwork = ((lwork)/32)*32;
    lwork *= sizeof(double);
    if ( lwork + 128*sizeof(double) < global_memory )
      lwork += (128*sizeof(double));
    cudaMalloc( (void **)&work, lwork );
  } // constructor

  // query methods
  int get_numbers_MP ( void )
  {
    return numbers_MPs;
  }

  void get_device_Name ( char *name )
  {
    strcpy(name, device_Name);
  }

  int get_device_Major ( void )
  {
    return device_Major;
  }

  int get_device_Minor ( void )
  {
    return device_Minor;
  }

  int get_driver_Version ( void )
  {
    return driver_Version;
  }

  int get_runtime_Version ( void )
  {
    return runtime_Version;
  }

  size_t get_device_Memory ( void )
  {
    return global_memory;
  }

  size_t get_device_WorkSize ( void )
  {
    return lwork;
  }

  void * get_device_WorkArea ( void )
  {
    return work;
  }
};


static ASPEN_GPU_info * GPU = NULL;

extern "C" int
API(init) ( int device_id )
{
  if ( GPU != NULL ) {
    fprintf( stderr, "ASPEN was already initialized.\n" );
    fprintf( stderr, "You might forgot to shutdown.\n" );
    return 1;
  }
  GPU = new ASPEN_GPU_info( device_id );
  return 0;
}

extern "C" int
API(shutdown) ( void )
{
  delete GPU;
  GPU = NULL;
  return 0;
}

#define	check_GPU( ... )						\
  _MACRO_WRAP_ (							\
		if ( GPU == NULL ) {					\
                  perror( "not Initialized" ); exit(1); };              \
                                                                        )


extern "C" int
API(get_numbers_MP) ( void )
{
  check_GPU( );
  return GPU->get_numbers_MP ( );
}

extern "C" void
API(get_device_Name) ( char * name )
{
  check_GPU( );
  GPU->get_device_Name ( name );
}

extern "C" int
API(get_device_Major) ( void )
{
  check_GPU( );
  return GPU->get_device_Major ( );
}

extern "C" int
API(get_device_Minor) ( void )
{
  check_GPU( );
  return GPU->get_device_Minor ( );
}

extern "C" int
API(get_device_CC) ( void )
{
  check_GPU( );
  return (GPU->get_device_Minor ( ) + GPU->get_device_Major ( ) * 10) * 10;
}

extern "C" int
API(get_driver_Version) ( void )
{
  check_GPU( );
  return GPU->get_driver_Version ( );
}

extern "C" int
API(get_runtime_Version) ( void )
{
  check_GPU( );
  return GPU->get_runtime_Version ( );
}

extern "C" size_t 
API(get_device_Memory) ( void )
{
  check_GPU( );
  return GPU->get_device_Memory ( );
}

extern "C" size_t 
API(get_device_WorkSize) ( void )
{
  check_GPU( );
  return GPU->get_device_WorkSize ( );
}

extern "C" void *
API(get_device_WorkArea) ( void )
{
  check_GPU( );
  return GPU->get_device_WorkArea ( );
}

