#include "aspen_devel.h"

int print_dev_info ( const char *device, const int MPs, const int CG, const size_t MMEM,
		     const size_t MAXDIM, const size_t MAXDIM2, const size_t MAXDIM3, const size_t MAXDIM4 );

int
main( int argc, char *argv[] )
{
  int	id = 0;
  if ( argc > 1 ) {
    sscanf( argv[1], "%d", &id );
  } else {
    char *id_str = getenv("ASPEN_GPU_ID");
    if ( id_str ) {
      sscanf( id_str, "%d", &id );
    }
  }

  cudaError_t err = cudaSetDevice( id );
  if ( err != cudaSuccess ) {
    perror("CUDA device open FAILED");
    print_dev_info ( "unknown", (int)0, (int)0, (size_t)0, (size_t)0, (size_t)0, (size_t)0, (size_t)0 );
    return EXIT_FAILURE;
  }

  ASPEN_init   ( id );

  char	device[1024]; ASPEN_get_device_Name( device );
  int	MPs     = ASPEN_get_numbers_MP( );
  int	CG      = ASPEN_get_device_CC( );
  size_t	MMEM    = ASPEN_get_device_Memory( );
  size_t	MAXDIM, MAXDIM2, MAXDIM3, MAXDIM4;
  MMEM -= ASPEN_get_device_WorkSize( );

  int i, L = strlen( device );
  for(i=0; i<L; i++) if ( device[i] == ' ' ) device[i] = '-';

  MAXDIM  = MMEM / (sizeof(double));
  MAXDIM  = (size_t)(sqrt((double)MAXDIM)*0.95);

  MAXDIM2 = MMEM / (sizeof(float));
  MAXDIM2 = (size_t)(sqrt((double)MAXDIM2)*0.95);

  MAXDIM3 = MMEM / (sizeof(double)*2);
  MAXDIM3 = (size_t)(sqrt((double)MAXDIM3)*0.95);

  MAXDIM4 = MMEM / (sizeof(double)*4);
  MAXDIM4 = (size_t)(sqrt((double)MAXDIM4)*0.95);

  print_dev_info ( device, MPs, CG, MMEM, MAXDIM, MAXDIM2, MAXDIM3, MAXDIM4 );

  printf( "CUDA= %d\n",  ASPEN_get_runtime_Version() );

  return EXIT_SUCCESS;
}

int
print_dev_info ( const char *device, const int MPs, const int CG, const size_t MMEM,
		 const size_t MAXDIM, const size_t MAXDIM2, const size_t MAXDIM3, const size_t MAXDIM4 )
{
  printf( "DEVICE= %s\n"   , device  );
  printf( "MP= %d\n"       , MPs     );
  printf( "CG= %d\n"       , CG      );
  printf( "MAXmem= %ld\n"  , MMEM    );
  printf( "MAXDIM= %ld\n"  , MAXDIM  );
  printf( "MAXDIM2= %ld\n" , MAXDIM2 );
  printf( "MAXDIM3= %ld\n" , MAXDIM3 );
  printf( "MAXDIM4= %ld\n" , MAXDIM4 );
}

