#include "aspen_devel.h"
#include "aspen_complex.h"


template < class TYPE, int BLOCK_SIZE, int MULTIPLE >
__global__ void 
API(SCAL_kernel) ( const TYPE alpha, TYPE * x, const long n, const long incx )
{
  long pos = threadIdx.x + MULTIPLE * BLOCK_SIZE * blockIdx.x;
  x += pos*incx;

  if ( pos+BLOCK_SIZE*MULTIPLE <= n ) {
#pragma unroll
    for ( int i=0; i<MULTIPLE; ++i ) {
      x[0] *= alpha;
      if ( MULTIPLE > 1 ) { x += BLOCK_SIZE*incx; }
    }
  } else {
#pragma unroll
    for ( int i=0; i<MULTIPLE; ++i ) {
      if ( pos+BLOCK_SIZE*i < n ) x[0] *= alpha;
      if ( MULTIPLE > 1 ) { x += BLOCK_SIZE*incx; }
    }
  }
}


template < class TYPE, int BLOCK_SIZE, int MULTIPLE >
__host__ void
API(SCAL_host) ( const TYPE alpha, TYPE * x, const int n, const int incx )
{
  if ( n <= 0 ) return;

  int t1 = BLOCK_SIZE;
  int tt = MULTIPLE*t1;
  int n_blk = n / tt;
  int n_res = ((n_blk*tt<n)?1:0);
  int g1 = n_blk + n_res;

  API(SCAL_kernel) < TYPE , BLOCK_SIZE, MULTIPLE >
    <<< g1, t1
    >>> ( alpha, x, (long) n, (long) incx );

}


#define BLOCK_SIZE	(16*12)
#define	MULTIPLE	(2)

extern "C" void
API(sscal) ( const int n, const float alpha, float * x, const int incx )
{
  API(SCAL_host) < float, BLOCK_SIZE, MULTIPLE > ( alpha, x, n, incx );
}

extern "C" void
API(dscal) ( const int n, double alpha, double * x, const int incx )
{
  API(SCAL_host) < double, BLOCK_SIZE, MULTIPLE > ( alpha, x, n, incx );
}

extern "C" void
API(cscal) ( const int n, const cuFloatComplex alpha, cuFloatComplex * x, const int incx )
{
  API(SCAL_host) < cuFloatComplex, BLOCK_SIZE, MULTIPLE > ( alpha, x, n, incx );
}

extern "C" void
API(zscal) ( const int n, const cuDoubleComplex alpha, cuDoubleComplex * x, const int incx )
{
  API(SCAL_host) < cuDoubleComplex, BLOCK_SIZE, MULTIPLE > ( alpha, x, n, incx );
}

