#include "aspen_devel.h"
#include "aspen_complex.h"


template < class TYPE, int BLOCK_SIZE, int MULTIPLE >
__global__ void 
API(SWAP_kernel) ( TYPE * y, TYPE * x, const long n, const long incx, const long incy )
{
  long pos = threadIdx.x + MULTIPLE * BLOCK_SIZE * blockIdx.x;
  x += pos*incx; y += pos*incy;

  if ( pos+BLOCK_SIZE*MULTIPLE <= n ) {
#pragma unroll
    for ( int i=0; i<MULTIPLE; ++i ) {
      TYPE t = x[0]; x[0] = y[0]; y[0] = t;
      if ( MULTIPLE > 1 ) { x += BLOCK_SIZE*incx; y += BLOCK_SIZE*incy; }
    }
  } else {
#pragma unroll
    for ( int i=0; i<MULTIPLE; ++i ) {
      if ( pos+BLOCK_SIZE*i < n ) { TYPE t = x[0]; x[0] = y[0]; y[0] = t; }
      if ( MULTIPLE > 1 ) { x += BLOCK_SIZE*incx; y += BLOCK_SIZE*incy; }
    }
  }
}


template < class TYPE, int BLOCK_SIZE, int MULTIPLE >
__host__ void
API(SWAP_host) ( TYPE * y, TYPE * x, const int n, const int incx, const int incy )
{
  if ( n <= 0 ) return;

  int t1 = BLOCK_SIZE;
  int tt = MULTIPLE*t1;
  int n_blk = n / tt;
  int n_res = ((n_blk*tt<n)?1:0);
  int g1 = n_blk + n_res;

  API(SWAP_kernel) < TYPE , BLOCK_SIZE, MULTIPLE >
    <<< g1, t1
    >>> ( y, x, (long) n, (long) incx, (long) incy );

}


#define BLOCK_SIZE	(16*12)
#define	MULTIPLE	(2)


extern "C" void
API(sswap) ( const int n, float * x, const int incx, float * y, const int incy )
{
  API(SWAP_host) < float, BLOCK_SIZE, MULTIPLE > ( y, x, n, incx, incy );
}

extern "C" void
API(dswap) ( const int n, double * x, const int incx, double * y, const int incy )
{
  API(SWAP_host) < double, BLOCK_SIZE, MULTIPLE > ( y, x, n, incx, incy );
}

extern "C" void
API(cswap) ( const int n, cuFloatComplex * x, const int incx, cuFloatComplex * y, const int incy )
{
  API(SWAP_host) < cuFloatComplex, BLOCK_SIZE, MULTIPLE > ( y, x, n, incx, incy );
}

extern "C" void
API(zswap) ( const int n, cuDoubleComplex * x, const int incx, cuDoubleComplex * y, const int incy )
{
  API(SWAP_host) < cuDoubleComplex, BLOCK_SIZE, MULTIPLE > ( y, x, n, incx, incy );
}


