#!/bin/bash

prolog()
{
	if [ x${CUDA_PATH} = x ]; then
		echo "Please set up the CUDA_PATH variable."
		echo "(1) export CUDA_PATH=/usr/local/cuda"
		echo "(2) setenv CUDA_PATH /usr/local/cuda"
		exit 1
	fi
	AT_WORK_DIR=logs
	AT_LOG_FILENAME=dgemvn-at-log
	AT_OPT_FILENAME=dgemvn-at-opt.c
	AT_OPT2_FILENAME=dgemvn-at-opt2.h
	AT_OPT_EXE=dgemvn-at-pred
	AT_GP_PLOT=aa.plot
	BENCHCODE=../bench/test2-dgemv-n
	IN_DATA=IN-a

	if [ x${HOST} = x ]; then
		HOST=`/bin/hostname`
	fi

if [ x${CUDA_PATH} != x ]; then
        export LD_LIBRARY_PATH=$CUDA_PATH/lib64:$LD_LIBRARY_PATH
fi

	cd ../src; make get_dev_info; cd ../tuning
	../src/get_dev_info $ASPEN_GPU_ID > DEV_INFO
	cd ../src; make get_mult get_mult32 get_mult48; cd ../tuning

	DEVICE=`awk '/DEVICE=/{ print $2 }' DEV_INFO`
	MP=`awk '/MP=/{ print $2 }' DEV_INFO`
	CG=`awk '/CG=/{ print $2 }' DEV_INFO`
	MAXDIM=`awk '/MAXDIM=/{ print $2 }' DEV_INFO`

	\rm DEV_INFO

	if [ -e CURRENT_GPU ]; then
	CG_=`awk '{print $3*1}' CURRENT_GPU`
	if [ $CG -ne $CG_ ]; then
		\rm CURRENT_GPU
		echo '#define CURRENT_GPU '$CG > CURRENT_GPU
	fi
	else
		\rm CURRENT_GPU
		echo '#define CURRENT_GPU '$CG > CURRENT_GPU
	fi

	if [ $CG -eq 130 ]; then
		MAX_REGS=32768
		MAX_BLK=8
		SHMEM_UNIT_SIZE=512
	fi
	if [ $CG -eq 200 ]; then
		MAX_REGS=32768
		MAX_BLK=8
		SHMEM_UNIT_SIZE=128
	fi
	if [ $CG -eq 300 ]; then
		MAX_REGS=65536
		MAX_BLK=16
		SHMEM_UNIT_SIZE=256
	fi
	if [ $CG -eq 350 ]; then
		MAX_REGS=65536
		MAX_BLK=16
		SHMEM_UNIT_SIZE=256
	fi
	if [ $CG -eq 500 ]; then
		MAX_REGS=65536
		MAX_BLK=32
		SHMEM_UNIT_SIZE=256
	fi
	if [ $CG -eq 520 ]; then
		MAX_REGS=65536
		MAX_BLK=32
		SHMEM_UNIT_SIZE=256
	fi

	CORES=`cat /proc/cpuinfo | awk 'BEGIN{c=1}/cores/{c=$4}END{print c}'`
	if [ $CORES -gt 1 ]; then
		EXEC='' #'/usr/bin/numactl --physcpubind=0 '
	else
		EXEC=''
	fi

	FIT_INIT='o0=10; o1=1; c0=1; c1=1; c2=1;' 

	ALL_BLOCK_X_LIST=" \
		100006 100007 100008 100009 100010 \
		100011 100012 100013 100014 100015 \
		100016 \
		100106 100107 100108 100109 \
 		1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 \
 		101 102 103 104 105 106 107 108 109 110 111 112 113 114 \
 		201 202 203 204 205 206 207 208 209 210 211 212 \
		1001 1002 1003 \
		1101 1102 1103 \
		2001 2002 2003 2004 2005 2006 2007 2008 \
		2101 2102 2103 \
		3001 \
		3101 \
		4001 4002 4003 4004 4005 4006 \
		5001 \
		6001 6002 6003 6004 \
		7001 \
		8001 8002 \
		9001 \
		10001 10002 "

	BLOCK_X_LIST=$ALL_BLOCK_X_LIST

	if [ x${DEVICE} = xunknown ]; then
		echo '**** CAUTION !!! ****'
		echo 'CUDA environment does not match to your GPU'
		echo 'Check the installed NVIDIA driver and the version of CUDA!'
		exit 1
	fi

	/bin/sh ./touch_header.sh

	if [ -e $AT_WORK_DIR ]; then
		\rm -rf $AT_WORK_DIR
	fi
}

echo_Message()
{
	echo '<--/****************************************'
	echo 'Automatic Performance Tuning for DGEMV -N'
	date
	echo -n 'Host on '; echo `hostname`
	echo 'Device is '$DEVICE
	if [ x${1} = xbegin ]; then
		echo 'Start.'
	fi
	if [ x${1} = xend ]; then
		echo 'Successfully completed.'
	fi
	if [ x${1} = xterminate ]; then
		echo 'Terminated unfortunately.'
	fi
	echo '****************************************/-->'
}

update_ASPEN()
{
	cd ../src
	\rm dgemv_n.cu_o
	\rm dgemv_n-*.cu_o
	make -j4
	cd ../bench
	\rm test.o test2.o
	make test-d
	cd ../tuning
}

init_IN_file()
{
	echo 1 > $IN_DATA
	for N in \
		50 100 150 200 250 300 350 400 450 500 550 600 650 700 750 \
		800 900 1000 1100 1200 1300 1400 1500 1600 1700 1900 \
		2000 2100 2300 2400 2500 2700 2900 3200 3400
	do
		echo  $N >> $IN_DATA
	done
	N=3700
        NN=$MAXDIM
#	NS=`expr 56 "*" $MP`
	NS=`expr 71 "*" $MP`
        while [ $N -lt $NN ]
        do
		echo $N >> $IN_DATA
                N=`expr $N "+" $NS`
	done
		echo $MAXDIM >> $IN_DATA	
	echo 0 >> $IN_DATA
}

init_AT_API()
{
	for BLOCK_X in $BLOCK_X_LIST
	do
		echo '#define KERNEL_'$BLOCK_X' 1'
	done
}

do_AT_sampler1()
{
cat << EOF_IN1 > IN-1
1
6000
-1
EOF_IN1

	echo '<--/****************************************'
	echo 'Check the number of registers and shared memory'
	echo 'and Screen Kernels by occupancy.'
	echo '****************************************/-->'

	\touch dgemvn_opt4.h
	\rm    dgemvn_opt4.h
	\touch dgemvn_opt5.h
	\rm    dgemvn_opt5.h

	echo '#if 0' > dgemvn_opt4.h
	echo_Message sampler1 >> dgemvn_opt4.h
cat << EOF_1 >> dgemvn_opt4.h
#endif
#ifdef	ALL_KERNELS
#undef	ALL_KERNELS
#endif
#define	ALL_KERNELS	1
EOF_1


	/bin/sh ./get_reguse.sh $CG GEMVN_2D_kernel \
		../src/dgemv_n-X_template \
		../src/dgemv_n-X_template \
		../src/dgemv_n-X_template \
		../src/dgemv_n-X_template \
		../src/dgemv_n-X_template \
		../src/dgemv_n-X_template \
		> log-reg-all


	for BLOCK_X in $BLOCK_X_LIST
	do

# ============================================================

#		BASE=`expr $BLOCK_X "/" 2000`
#		BASE=`expr $BASE "+" 1`
#		BASE=`expr $BASE "*" 32`
		if [ $BLOCK_X -lt 100000 ]; then
		BASE=`expr $BLOCK_X "/" 1000`
		BASE=`expr $BASE "+" 2`
		BASE=`expr $BASE "*" 16`
		else
		BASE=16
		fi
		BLOCK_Y=`expr $BLOCK_X "%" 100`

		cat log-reg-all | awk '{if(($2=='$BASE')&&($3=='$BLOCK_Y'))print $1}' > CX
		CX=`cat CX`
		if [ x${CX} = "x" ]; then
			CX=1
		fi

# ============================================================

		M1=`expr $BLOCK_Y "*" $BASE`

		M=`expr $BLOCK_Y "/" 2`
		M=`expr $M "*" 2`
		M=`expr $M "+" 1`
		M=`expr $M "*" $BASE`
		M2=`expr $M "*" 8`

		M=`expr $BLOCK_X "/" 100`
		M3=`expr $M "%" 10`

if [ $CG -lt 200 ]; then
		MM_PAT=`../../src/get_mult $M1 $CX $M2 $CG`
else
		MM_PAT=`../../src/get_mult48 $M1 $CX $M2 $CG`
fi

		M=`expr $M3 "+" 1`
		\touch log_MM
		\rm log_MM
		for MM in $MM_PAT
		do
			echo $MM >> log_MM
		done
		MM=`cat log_MM | \
			tail -$M | head -1 `
		MULTIPLICITY=`echo $MM | \
			awk '{ gsub(/\(.*\)/,""); print; }'`
		OO=`echo $MM | \
			awk '{ gsub(/[\(\)]/," "); if($2==100)$2=99; print $2; }'`
		\rm log_MM

		M=`expr $MP "*" $BASE`
		CYCLE=`expr $M "*" $MULTIPLICITY`

cat << EOF_1 >> dgemvn_opt4.h

// ID[$BLOCK_X] ($BASE,$BLOCK_Y) * $MULTIPLICITY
#define	_BLOCKSIZE_$BLOCK_X	$BASE
#define	_UX_$BLOCK_X		$BLOCK_Y
#define	_MX_$BLOCK_X		$M2
#define	_CX_$BLOCK_X		$CX
#define	_MULTI_$BLOCK_X		$MULTIPLICITY
#define	_OCCUP_$BLOCK_X		$OO
#define	_CYCLE_$BLOCK_X		$CYCLE
EOF_1

# ============================================================

		echo -e 'Kernel ID:\t'$BLOCK_X'\tRegs:\t'$CX'\tMULTI:\t'$MULTIPLICITY'\tCYCLE:\t'$CYCLE'\t/ '$OO | tee -a dgemvn_opt5.h

	done

cat << EOF_2 >> dgemvn_opt4.h

#if defined(__CY__)
#undef	__CY__
#endif
#define	__CY__
EOF_2

	if [ $CG -le 200 ]; then
		O_LIMIT=16
	else
		O_LIMIT=20
	fi

	cat dgemvn_opt5.h | \
		sort -r -k 11 | sort -r -k 9 | \
		awk 'BEGIN{ \
			; \
		} \
		{ \
			multi = $7; \
			cycle = $9; \
			occup = $11; \
			if ( cycle != CYC ) { \
				CYC = cycle; \
				OO = occup; \
				NN = 0; \
			} \
			if ( occup != OO ) { \
				if ( NN == 0 ) { \
					NN = occup; \
				} \
			} \
			if ( multi < 2 || occup < '$O_LIMIT' || \
				( occup != OO && occup != NN && \
					OO*.75 > occup ) ) { \
				print $3; \
			} \
		} \
		END{ \
			;
		}' | \
		tee skip_list | \
		awk '{ print "Sampling for Kernel "$1" will be skipped" }'

	cat skip_list | \
		awk 'BEGIN{ \
			print ""; \
		} \
		{ \
			print "#undef _OCCUP_"$1; \
			print "#define _OCCUP_"$1" 0"; \
		} \
		END{ \
			print ""; \
		}' >> dgemvn_opt4.h

	echo '*/' >> dgemvn_opt5.h
	cat dgemvn_opt4.h >> dgemvn_opt5.h

	echo '/*' > dgemvn_opt4.h
	cat dgemvn_opt5.h >> dgemvn_opt4.h

	\rm dgemvn_opt5.h

	\rm IN-1

}

do_AT_sampler()
{
	if [ ! -d $AT_WORK_DIR ]; then
		mkdir $AT_WORK_DIR
	fi

	echo '<--/****************************************'
	echo 'Enter the Sampling Phase, it may take a while.'
	echo '****************************************/-->'

	for BLOCK_X in $BLOCK_X_LIST
	do

#		BASE=`expr $BLOCK_X "/" 2000`
#		BASE=`expr $BASE "+" 1`
#		BASE=`expr $BASE "*" 32`
		if [ $BLOCK_X -lt 100000 ]; then
		BASE=`expr $BLOCK_X "/" 1000`
		BASE=`expr $BASE "+" 2`
		BASE=`expr $BASE "*" 16`
		else
		BASE=16
		fi
		BLOCK_Y=`expr $BLOCK_X "%" 100`

		cat dgemvn_opt4.h | \
		awk '/define[\t ]+_CYCLE_'$BLOCK_X'[\t ]+/{ \
			gsub(/\(/,""); gsub(/\)/,""); \
			if($3*1>=0){ print $3; exit } \
		}' > CY
		CY=`cat CY`

		cat dgemvn_opt4.h | \
		awk '/define[\t ]+_OCCUP_'$BLOCK_X'[\t ]+/{ \
			gsub(/\(/,""); gsub(/\)/,""); \
			if($3*1>=0){ print $3 } \
		}' | tail -1 > OO
		OO=`cat OO`


		echo 'Sampling for Kernel '$BLOCK_X
		echo 'CYCLE= '$CY

		touch $IN_DATA--a
		\rm $IN_DATA--a

		cat $IN_DATA | \
			awk '/^[0-9]/{ \
				if ( $1 < 1 ) { \
					N=CY='$CY'; \
					MAXDIM='$MAXDIM'; \
					count=0; \
					for(N=CY;N-32<'$MAXDIM';N+=CY) { \
					count++; \
			if ( count < 8 ) { \
				if(N-32<=MAXDIM)print N-32; \
			} \
			if ( count < 12 ) { \
				if(N-16<=MAXDIM)print N-16; \
			} \
			if ( count < 16 ) { \
				if(N-04<=MAXDIM)print N-04; \
			} \
			if ( count < 20 ) { \
				if(N-01<=MAXDIM)print N-01; \
				if(N   <=MAXDIM)print N   ; \
			} \
			if ( count < 16 ) { \
				if(N+04<=MAXDIM)print N+04; \
			} \
			if ( count < 12 ) { \
				if(N+16<=MAXDIM)print N+16; \
			} \
			if ( count < 8 ) { \
				if(N+32<=MAXDIM)print N+32; \
			} \
					} \
					exit;
				} \
			} \
			{ print; }' > $IN_DATA--a

		cat $IN_DATA--a | \
			awk '{ printf("%06d\n", $0); }' | \
			sort | \
			uniq | \
			awk '{ print $0*1; } END{ print 0; }' > $IN_DATA--b
			

		if [ $CG -le 200 ]; then
			O_LIMIT=16
		else
			O_LIMIT=20
		fi

		if [ $OO -lt $O_LIMIT ]; then

		echo 'Kernel ID: '$BLOCK_X' is skipped due to the Low Occupancy'
		cat $IN_DATA--b | \
			awk 'BEGIN{ getline } \
				{ if ( $1 < 1 ) exit; \
					print "N= "$1" 0 [s] 0 GFLOPS 0 0" } \
			' > $AT_WORK_DIR/log-$BLOCK_X

		else

		$EXEC $BENCHCODE $IN_DATA--b $BLOCK_X | \
			awk '/^N=/{print; fflush(); }' | \
			tee $AT_WORK_DIR/log-$BLOCK_X

		fi

		cp $IN_DATA--b $AT_WORK_DIR/$IN_DATA--$BLOCK_X

		\rm $IN_DATA--a $IN_DATA--b


		echo 'CYCLE= '$CY | tee -a logs/log-$BLOCK_X

	done
}

epilog()
{
	if [ -d $AT_WORK_DIR ]; then
		LOG=log-dgemvn-AT1X.`date +%Y.%m.%d.%T`
		mv $AT_WORK_DIR $LOG
		cp ../src/dgemv_n-X_template.cu $LOG
		cp ../template//gemv_n-small_template.incl $LOG
		mv $AT_LOG_FILENAME $LOG
	fi
	if [ -f $IN_DATA ]; then
		\rm $IN_DATA
	fi
	if [ -f $AT_OPT_EXE ]; then
        	\rm $AT_OPT_EXE
	fi
	if [ -f fit.log ]; then
		\rm fit.log
	fi
	if [ -f SdkMasterLog.csv ]; then
		\rm SdkMasterLog.csv
	fi
	if [ -f deviceQuery.txt ]; then
		\rm deviceQuery.txt
	fi
}



prolog

echo_Message begin

if [ x${CUDA_CHECK} = xFAILED ]; then
	echo '*** WARNING ***'
	echo 'CUDA_SDK_PATH does not match the one specified on dgemvn.sh'
	echo 'PLEASE correct CUDA_SDK_PATH.'
	exit 1
fi

if [ $CG -lt 200 ]; then

	touch $AT_LOG_FILENAME

	generate_AT_API > $AT_OPT_FILENAME

	update_ASPEN

	echo '*** CAUTION ***'
	echo 'Because GT200 core uses a propriately tuned routine for DGEMV-N,'
	echo 'NO tuning process is required at the moment.'

	epilog

	echo_Message end

	exit 0;
fi

init_IN_file

init_AT_API >$AT_OPT2_FILENAME

touch  dgemvn_opt4.h
\rm  dgemvn_opt4.h
touch  dgemvn_opt4.h

touch dgemvn-auto2.h
\rm dgemvn-auto2.h
touch dgemvn-auto2.h

update_ASPEN

do_AT_sampler1

update_ASPEN

do_AT_sampler

cd  logs/
touch dgemvn-auto2.h
\rm  dgemvn-auto2.h
/bin/sh  ../bmvnd.sh
/bin/sh  ../cmvnd.sh

cat dgemvn-auto.h

/bin/sh  ../d_filter.sh dgemvn-auto.h dgemvn-auto2.h

echo '#if 0'   > dgemvn-auto_.h
echo_Message  >> dgemvn-auto_.h
echo '#endif' >> dgemvn-auto_.h
cat dgemvn-auto2.h >> dgemvn-auto_.h
mv dgemvn-auto_.h dgemvn-auto2.h

cat dgemvn-auto2.h

cp  dgemvn-auto2.h ../
cd  ../

update_ASPEN

epilog

echo_Message end

touch .done-dgemvn

exit 0


